@@ -1,27 +1,68 @@
-# Copyright (C) 2019  The Software Heritage developers
+# Copyright (C) 2019-2020  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
-import aiohttp
-import click
 import logging
-import shutil
-import sys
-
 from pathlib import Path
-from typing import Any, Dict, Tuple
+import sys
+from typing import TYPE_CHECKING, Any, Dict, Set, Tuple
 
-import swh.model.exceptions
+# WARNING: do not import unnecessary things here to keep cli startup time under
+# control
+import click
 
-from swh.core import config
 from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup
-from swh.graph import client, webgraph
-from swh.graph.backend import NODE2PID_EXT, PID2NODE_EXT
-from swh.graph.pid import PidToNodeMap, NodeToPidMap
-from swh.graph.server.app import make_app
-from swh.graph.backend import Backend
-from swh.model.identifiers import parse_persistent_identifier
+from swh.core.cli import swh as swh_cli_group
+
+if TYPE_CHECKING:
+    from swh.graph.webgraph import CompressionStep  # noqa
+
+
+class StepOption(click.ParamType):
+    """click type for specifying a compression step on the CLI
+
+    parse either individual steps, specified as step names or integers, or step
+    ranges
+
+    """
+
+    name = "compression step"
+
+    def convert(self, value, param, ctx):  # type: (...) -> Set[CompressionStep]
+        from swh.graph.webgraph import COMP_SEQ, CompressionStep  # noqa
+
+        steps: Set[CompressionStep] = set()
+
+        specs = value.split(",")
+        for spec in specs:
+            if "-" in spec:  # step range
+                (raw_l, raw_r) = spec.split("-", maxsplit=1)
+                if raw_l == "":  # no left endpoint
+                    raw_l = COMP_SEQ[0].name
+                if raw_r == "":  # no right endpoint
+                    raw_r = COMP_SEQ[-1].name
+                l_step = self.convert(raw_l, param, ctx)
+                r_step = self.convert(raw_r, param, ctx)
+                if len(l_step) != 1 or len(r_step) != 1:
+                    self.fail(f"invalid step specification: {value}, " f"see --help")
+                l_idx = l_step.pop()
+                r_idx = r_step.pop()
+                steps = steps.union(
+                    set(CompressionStep(i) for i in range(l_idx.value, r_idx.value + 1))
+                )
+            else:  # singleton step
+                try:
+                    steps.add(CompressionStep(int(spec)))  # integer step
+                except ValueError:
+                    try:
+                        steps.add(CompressionStep[spec.upper()])  # step name
+                    except KeyError:
+                        self.fail(
+                            f"invalid step specification: {value}, " f"see --help"
+                        )
+
+        return steps
 
 
 class PathlibPath(click.Path):
@@ -34,7 +75,7 @@ class PathlibPath(click.Path):
 DEFAULT_CONFIG: Dict[str, Tuple[str, Any]] = {"graph": ("dict", {})}
 
 
-@click.group(name="graph", context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)
+@swh_cli_group.group(name="graph", context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)
 @click.option(
     "--config-file",
     "-C",
@@ -43,10 +84,11 @@ DEFAULT_CONFIG: Dict[str, Tuple[str, Any]] = {"graph": ("dict", {})}
     help="YAML configuration file",
 )
 @click.pass_context
-def cli(ctx, config_file):
+def graph_cli_group(ctx, config_file):
     """Software Heritage graph tools."""
-    ctx.ensure_object(dict)
+    from swh.core import config
 
+    ctx.ensure_object(dict)
     conf = config.read(config_file, DEFAULT_CONFIG)
     if "graph" not in conf:
         raise ValueError(
@@ -55,12 +97,14 @@ def cli(ctx, config_file):
     ctx.obj["config"] = conf
 
 
-@cli.command("api-client")
+@graph_cli_group.command("api-client")
 @click.option("--host", default="localhost", help="Graph server host")
 @click.option("--port", default="5009", help="Graph server port")
 @click.pass_context
 def api_client(ctx, host, port):
-    """client for the graph REST service"""
+    """client for the graph RPC service"""
+    from swh.graph import client
+
     url = "http://{}:{}".format(host, port)
     app = client.RemoteGraphClient(url)
 
@@ -68,44 +112,52 @@ def api_client(ctx, host, port):
     print(app.stats())
 
 
-@cli.group("map")
+@graph_cli_group.group("map")
 @click.pass_context
 def map(ctx):
     """Manage swh-graph on-disk maps"""
     pass
 
 
-def dump_pid2node(filename):
-    for (pid, int) in PidToNodeMap(filename):
-        print("{}\t{}".format(pid, int))
+def dump_swhid2node(filename):
+    from swh.graph.swhid import SwhidToNodeMap
+
+    for (swhid, int) in SwhidToNodeMap(filename):
+        print("{}\t{}".format(swhid, int))
 
 
-def dump_node2pid(filename):
-    for (int, pid) in NodeToPidMap(filename):
-        print("{}\t{}".format(int, pid))
+def dump_node2swhid(filename):
+    from swh.graph.swhid import NodeToSwhidMap
+
+    for (int, swhid) in NodeToSwhidMap(filename):
+        print("{}\t{}".format(int, swhid))
 
 
-def restore_pid2node(filename):
-    """read a textual PID->int map from stdin and write its binary version to
+def restore_swhid2node(filename):
+    """read a textual SWHID->int map from stdin and write its binary version to
     filename
 
     """
+    from swh.graph.swhid import SwhidToNodeMap
+
     with open(filename, "wb") as dst:
         for line in sys.stdin:
-            (str_pid, str_int) = line.split()
-            PidToNodeMap.write_record(dst, str_pid, int(str_int))
+            (str_swhid, str_int) = line.split()
+            SwhidToNodeMap.write_record(dst, str_swhid, int(str_int))
 
 
-def restore_node2pid(filename, length):
-    """read a textual int->PID map from stdin and write its binary version to
+def restore_node2swhid(filename, length):
+    """read a textual int->SWHID map from stdin and write its binary version to
     filename
 
     """
-    node2pid = NodeToPidMap(filename, mode="wb", length=length)
+    from swh.graph.swhid import NodeToSwhidMap
+
+    node2swhid = NodeToSwhidMap(filename, mode="wb", length=length)
     for line in sys.stdin:
-        (str_int, str_pid) = line.split()
-        node2pid[int(str_int)] = str_pid
-    node2pid.close()
+        (str_int, str_swhid) = line.split()
+        node2swhid[int(str_int)] = str_swhid
+    node2swhid.close()
 
 
 @map.command("dump")
@@ -114,17 +166,17 @@ def restore_node2pid(filename, length):
     "-t",
     "map_type",
     required=True,
-    type=click.Choice(["pid2node", "node2pid"]),
+    type=click.Choice(["swhid2node", "node2swhid"]),
     help="type of map to dump",
 )
 @click.argument("filename", required=True, type=click.Path(exists=True))
 @click.pass_context
 def dump_map(ctx, map_type, filename):
-    """Dump a binary PID<->node map to textual format."""
-    if map_type == "pid2node":
-        dump_pid2node(filename)
-    elif map_type == "node2pid":
-        dump_node2pid(filename)
+    """Dump a binary SWHID<->node map to textual format."""
+    if map_type == "swhid2node":
+        dump_swhid2node(filename)
+    elif map_type == "node2swhid":
+        dump_node2swhid(filename)
     else:
         raise ValueError("invalid map type: " + map_type)
     pass
@@ -136,7 +188,7 @@ def dump_map(ctx, map_type, filename):
     "-t",
     "map_type",
     required=True,
-    type=click.Choice(["pid2node", "node2pid"]),
+    type=click.Choice(["swhid2node", "node2swhid"]),
     help="type of map to dump",
 )
 @click.option(
@@ -144,20 +196,20 @@ def dump_map(ctx, map_type, filename):
     "-l",
     type=int,
     help="""map size in number of logical records
-              (required for node2pid maps)""",
+              (required for node2swhid maps)""",
 )
 @click.argument("filename", required=True, type=click.Path())
 @click.pass_context
 def restore_map(ctx, map_type, length, filename):
-    """Restore a binary PID<->node map from textual format."""
-    if map_type == "pid2node":
-        restore_pid2node(filename)
-    elif map_type == "node2pid":
+    """Restore a binary SWHID<->node map from textual format."""
+    if map_type == "swhid2node":
+        restore_swhid2node(filename)
+    elif map_type == "node2swhid":
         if length is None:
             raise click.UsageError(
                 "map length is required when restoring {} maps".format(map_type), ctx
             )
-        restore_node2pid(filename, length)
+        restore_node2swhid(filename, length)
     else:
         raise ValueError("invalid map type: " + map_type)
 
@@ -168,7 +220,7 @@ def restore_map(ctx, map_type, length, filename):
     "-t",
     "map_type",
     required=True,
-    type=click.Choice(["pid2node", "node2pid"]),
+    type=click.Choice(["swhid2node", "node2swhid"]),
     help="type of map to write",
 )
 @click.argument("filename", required=True, type=click.Path())
@@ -176,23 +228,25 @@ def restore_map(ctx, map_type, length, filename):
 def write(ctx, map_type, filename):
     """Write a map to disk sequentially.
 
-    read from stdin a textual PID->node mapping (for pid2node, or a simple
-    sequence of PIDs for node2pid) and write it to disk in the requested binary
+    read from stdin a textual SWHID->node mapping (for swhid2node, or a simple
+    sequence of SWHIDs for node2swhid) and write it to disk in the requested binary
     map format
 
     note that no sorting is applied, so the input should already be sorted as
-    required by the chosen map type (by PID for pid2node, by int for node2pid)
+    required by the chosen map type (by SWHID for swhid2node, by int for node2swhid)
 
     """
+    from swh.graph.swhid import NodeToSwhidMap, SwhidToNodeMap
+
     with open(filename, "wb") as f:
-        if map_type == "pid2node":
+        if map_type == "swhid2node":
             for line in sys.stdin:
-                (pid, int_str) = line.rstrip().split(maxsplit=1)
-                PidToNodeMap.write_record(f, pid, int(int_str))
-        elif map_type == "node2pid":
+                (swhid, int_str) = line.rstrip().split(maxsplit=1)
+                SwhidToNodeMap.write_record(f, swhid, int(int_str))
+        elif map_type == "node2swhid":
             for line in sys.stdin:
-                pid = line.rstrip()
-                NodeToPidMap.write_record(f, pid)
+                swhid = line.rstrip()
+                NodeToSwhidMap.write_record(f, swhid)
         else:
             raise ValueError("invalid map type: " + map_type)
 
@@ -205,9 +259,9 @@ def write(ctx, map_type, filename):
 def map_lookup(graph, identifiers):
     """Lookup identifiers using on-disk maps.
 
-    Depending on the identifier type lookup either a PID into a PID->node (and
+    Depending on the identifier type lookup either a SWHID into a SWHID->node (and
     return the node integer identifier) or, vice-versa, lookup a node integer
-    identifier into a node->PID (and return the PID).  The desired behavior is
+    identifier into a node->SWHID (and return the SWHID).  The desired behavior is
     chosen depending on the syntax of each given identifier.
 
     Identifiers can be passed either directly on the command line or on
@@ -215,29 +269,34 @@ def map_lookup(graph, identifiers):
     readline()) in stdin will be preserved in stdout.
 
     """
+    from swh.graph.backend import NODE2SWHID_EXT, SWHID2NODE_EXT
+    from swh.graph.swhid import NodeToSwhidMap, SwhidToNodeMap
+    import swh.model.exceptions
+    from swh.model.identifiers import ExtendedSWHID
+
     success = True  # no identifiers failed to be looked up
-    pid2node = PidToNodeMap(f"{graph}.{PID2NODE_EXT}")
-    node2pid = NodeToPidMap(f"{graph}.{NODE2PID_EXT}")
+    swhid2node = SwhidToNodeMap(f"{graph}.{SWHID2NODE_EXT}")
+    node2swhid = NodeToSwhidMap(f"{graph}.{NODE2SWHID_EXT}")
 
     def lookup(identifier):
-        nonlocal success, pid2node, node2pid
-        is_pid = None
+        nonlocal success, swhid2node, node2swhid
+        is_swhid = None
         try:
             int(identifier)
-            is_pid = False
+            is_swhid = False
         except ValueError:
             try:
-                parse_persistent_identifier(identifier)
-                is_pid = True
+                ExtendedSWHID.from_string(identifier)
+                is_swhid = True
             except swh.model.exceptions.ValidationError:
                 success = False
                 logging.error(f'invalid identifier: "{identifier}", skipping')
 
         try:
-            if is_pid:
-                return str(pid2node[identifier])
+            if is_swhid:
+                return str(swhid2node[identifier])
             else:
-                return node2pid[int(identifier)]
+                return node2swhid[int(identifier)]
         except KeyError:
             success = False
             logging.error(f'identifier not found: "{identifier}", skipping')
@@ -254,7 +313,7 @@ def map_lookup(graph, identifiers):
     sys.exit(0 if success else 1)
 
 
-@cli.command(name="rpc-serve")
+@graph_cli_group.command(name="rpc-serve")
 @click.option(
     "--host",
     "-h",
@@ -277,7 +336,12 @@ def map_lookup(graph, identifiers):
 )
 @click.pass_context
 def serve(ctx, host, port, graph):
-    """run the graph REST service"""
+    """run the graph RPC service"""
+    import aiohttp
+
+    from swh.graph.backend import Backend
+    from swh.graph.server.app import make_app
+
     backend = Backend(graph_path=graph, config=ctx.obj["config"])
     app = make_app(backend=backend)
 
@@ -285,7 +349,7 @@ def serve(ctx, host, port, graph):
         aiohttp.web.run_app(app, host=host, port=port)
 
 
-@cli.command()
+@graph_cli_group.command()
 @click.option(
     "--graph",
     "-g",
@@ -307,7 +371,7 @@ def serve(ctx, host, port, graph):
     "--steps",
     "-s",
     metavar="STEPS",
-    type=webgraph.StepOption(),
+    type=StepOption(),
     help="run only these compression steps (default: all steps)",
 )
 @click.pass_context
@@ -325,6 +389,8 @@ def compress(ctx, graph, out_dir, steps):
     also supported.
 
     """
+    from swh.graph import webgraph
+
     graph_name = graph.name
     in_dir = graph.parent
     try:
@@ -335,7 +401,7 @@ def compress(ctx, graph, out_dir, steps):
     webgraph.compress(graph_name, in_dir, out_dir, steps, conf)
 
 
-@cli.command(name="cachemount")
+@graph_cli_group.command(name="cachemount")
 @click.option(
     "--graph", "-g", required=True, metavar="GRAPH", help="compressed graph basename"
 )
@@ -354,12 +420,14 @@ def cachemount(ctx, graph, cache):
 
     This command creates a new directory at the path given by CACHE that has
     the same structure as the compressed graph basename, except it copies the
-    files that require mmap access (*.graph) but uses symlinks from the source
-    for all the other files (.map, .bin, ...).
+    files that require mmap access (:file:`{*}.graph`) but uses symlinks from the source
+    for all the other files (:file:`{*}.map`, :file:`{*}.bin`, ...).
 
     The command outputs the path to the memory cache directory (particularly
     useful when relying on the default value).
     """
+    import shutil
+
     cache.mkdir(parents=True)
     for src in Path(graph).parent.glob("*"):
         dst = cache / src.name
@@ -371,7 +439,7 @@ def cachemount(ctx, graph, cache):
 
 
 def main():
-    return cli(auto_envvar_prefix="SWH_GRAPH")
+    return graph_cli_group(auto_envvar_prefix="SWH_GRAPH")
 
 
 if __name__ == "__main__":
